using System;
using System.Collections;
using System.Reflection;
using System.Reflection.Emit;
using Micro.Interfaces;
using Micro.Interfaces.SymTab;

namespace Micro.Implementation
{
    public class Parser : IParser
    {
        private static readonly IList fFirstStat = new ArrayList();
        private static readonly IList fCondOpt = new ArrayList();
        private IScanner fScanner;
        private ICodeGen fCodeGen;
        private IToken fToken;
        private IToken fLa;
        private IError fError;
        private ISymTab fSymTab;


        public Parser(IScanner scanner, IError err, ICodeGen code)
        {
            fScanner = scanner;
            fError = err;
            fCodeGen = code;
            fSymTab = new SymTab(err);

            fFirstStat.Add(TokenKind.ident);
            fFirstStat.Add(TokenKind.callKW);
            fFirstStat.Add(TokenKind.ifKW);

            fCondOpt.Add(TokenKind.eql);
            fCondOpt.Add(TokenKind.lss);
            fCondOpt.Add(TokenKind.gtr);
            fCondOpt.Add(TokenKind.neq);
        }

        public void Parse()
        {
            // Initialize pre defined functions.
            IStruct intType = fSymTab.FindType("int").Type;
            IStruct charType = fSymTab.FindType("char").Type;
            
            ISymbol putiSym = fSymTab.Insert(SymbolKind.Func, "put_i", intType);
            fSymTab.OpenScope();
            fSymTab.Insert(SymbolKind.Arg, "i", intType);
            fSymTab.SetLocals(putiSym);
            BuildPutI(putiSym);
            fSymTab.CloseScope();

            ISymbol putcSym = fSymTab.Insert(SymbolKind.Func, "put_c", intType);
            fSymTab.OpenScope();
            fSymTab.Insert(SymbolKind.Arg, "c", charType);
            fSymTab.SetLocals(putcSym);
            BuildPutC(putcSym);
            fSymTab.CloseScope();

            ISymbol getiSym = fSymTab.Insert(SymbolKind.Func, "get_i", intType);
            BuildGetI(getiSym);

            ISymbol getcSym = fSymTab.Insert(SymbolKind.Func, "get_c", charType);
            BuildGetC(getcSym);

            // Start working
            Scan();
            Micro();
        }

        private void Micro()
        {
            // TODO add methods for each production
        }

        #region private auxillary methods

        private void Error(string msg)
        {
            fError.Error(fLa, msg);
        }

        private void Check(TokenKind expected)
        {
            if (fLa.Kind == expected)
            {
                Scan();
            }
            else
            {
                Error("unexpected token, expected: " + expected + ", found: " + fLa.Kind);
            }
        }

        private void Scan()
        {
            fToken = fLa;
            fLa = fScanner.Next();
        }

        private void BuildGetC(ISymbol getc)
        {
            // char get_c () {
            fCodeGen.CreateMetadata(getc);

            //   return (char) System.Console.Read();
            fCodeGen.EmitCall(typeof (Console).GetMethod("Read", new Type[0]));
            fCodeGen.Emit(OpCodes.Conv_U2);

            // }
            fCodeGen.Emit(OpCodes.Ret);
        }

        private void BuildGetI(ISymbol geti)
        {
            // int get_i ()
            fCodeGen.CreateMetadata(geti);
            // return Int32.Parse(Console.ReadLine());
            fCodeGen.EmitCall(
                typeof (Console).GetMethod("ReadLine", new Type[0]));
            fCodeGen.EmitCall(
                typeof (Int32).GetMethod("Parse", new Type[] {typeof (string)}));
            // }
            fCodeGen.Emit(OpCodes.Ret);
        }

        private void BuildPutC(ISymbol putc)
        {
            // int put_c (char c)
            fCodeGen.CreateMetadata(putc);

            // System.Console.Write(c)
            fCodeGen.Emit(OpCodes.Ldarg_0);
            MethodInfo writeLine =
                typeof (Console).GetMethod("WriteLine",
                                           new Type[] {typeof (char)}, null);
            fCodeGen.EmitCall(writeLine);

            // return 0;
            fCodeGen.Emit(OpCodes.Ldc_I4_0);
            // }
            fCodeGen.Emit(OpCodes.Ret);
        }

        private void BuildPutI(ISymbol puti)
        {
            // int put_i (int x)
            fCodeGen.CreateMetadata(puti);

            fCodeGen.Emit(OpCodes.Ldarg_0);
            MethodInfo writeLine =
                typeof (Console).GetMethod("WriteLine",
                                           new Type[] {typeof (int)}, null);
            fCodeGen.EmitCall(writeLine);

            // return 0;
            fCodeGen.Emit(OpCodes.Ldc_I4_0);
            // }
            fCodeGen.Emit(OpCodes.Ret);
        }

        #endregion
    }
}