using System.Reflection;
using System.Reflection.Emit;
using Micro.Interfaces.SymTab;

namespace Micro.Interfaces
{
    public interface ICodeGen
    {
        /// <summary>
        /// Creates the required metadata builder objects for the given Symbol.
        /// Call this after you inserted your Symbol into the symbol table.
        /// </summary>
        /// <param name="sym"></param>
        void CreateMetadata(ISymbol sym);

        /// <summary>
        /// Load the operand x onto the expression stack.
        /// </summary>
        /// <param name="x"></param>
        void Load(IItem x);

        /// <summary>
        /// Generate an assignment x = y.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        void Assign(IItem x, IItem y);

        /// <summary>
        /// emit a jump (condition) to the given label
        /// </summary>
        /// <param name="condition"></param>
        /// <param name="label"></param>
        void Jump(OpCode condition, Label label);

        /// <summary>
        /// True Jump. Generates conditional branch instruction.
        /// </summary>
        /// <param name="x"></param>
        void Jump(IItem x);

        /// <summary>
        /// Emit an opcode as is.
        /// </summary>
        /// <param name="opcode"></param>
        void Emit(OpCode opcode);

        /// <summary>
        /// Emit a call to the given method
        /// </summary>
        /// <param name="meth"></param>
        void EmitCall(MethodInfo meth);

        /// <summary>
        /// Comfort function to emit a call to the method represented
        /// by the given symbol.
        /// </summary>
        /// <param name="function"></param>
        void EmitCall(ISymbol function);

        /// <summary>
        /// Create a constant item.
        /// </summary>
        /// <param name="val"></param>
        /// <param name="type"></param>
        /// <returns>the new item</returns>
        IItem CreateItem(int val, IStruct type);

        /// <summary>
        /// Create an item from a symbol. Allowed symbol kinds are
        /// argument, local variable, function and constant.
        /// </summary>
        /// <param name="sym"></param>
        /// <returns>the created item</returns>
        IItem CreateItem(ISymbol sym);

        /// <summary>
        /// Create a new item from a relational operator.
        /// </summary>
        /// <param name="relop"></param>
        /// <returns>the new item</returns>
        IItem CreateItem(TokenKind relop);

        /// <summary>
        /// Create a new label as target for a jump.
        /// </summary>
        /// <returns>the new label</returns>
        Label CreateLabel();

        /// <summary>
        /// Mark a label, means fixup all the usages. Every label must
        /// be marked at some time to become usefull.
        /// </summary>
        /// <param name="label"></param>
        void MarkLabel(Label label);

        /// <summary>
        /// Generate an executable .NET-PE-File.
        /// </summary>
        void WritePEFile();
    }
}